# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio-Circulation example application.

SPHINX-START

First install Invenio-Circulation, setup the application and load
fixture data by running:

.. code-block:: console

   $ pip install -e .[all]
   $ cd examples
   $ ./app-setup.sh
   $ FLASK_APP=app.py flask fixtures loans

Next, start the development server:

.. code-block:: console

   $ export FLASK_APP=app.py FLASK_ENV=development
   $ flask run

and open the example application in your browser:

.. code-block:: console

    $ open http://127.0.0.1:5000/circulation/loan/1

To reset the example application run:

.. code-block:: console

    $ ./app-teardown.sh

SPHINX-END
"""

from __future__ import absolute_import, print_function

import os
import uuid

import click
from flask import Flask
from invenio_db import db
from invenio_db.ext import InvenioDB
from invenio_jsonschemas import InvenioJSONSchemas
from invenio_pidstore.ext import InvenioPIDStore
from invenio_records.ext import InvenioRecords
from invenio_records_rest.ext import InvenioRecordsREST
from invenio_records_rest.utils import PIDConverter
from invenio_records_rest.views import create_blueprint_from_app
from invenio_rest import InvenioREST
from invenio_search.ext import InvenioSearch

from invenio_circulation.api import Loan
from invenio_circulation.config import CIRCULATION_REST_ENDPOINTS
from invenio_circulation.ext import InvenioCirculation
from invenio_circulation.pidstore.minters import loan_pid_minter
from invenio_circulation.views import create_loan_actions_blueprint, \
    create_loan_for_item_blueprint, create_loan_replace_item_blueprint

# Create Flask application
app = Flask(__name__)
app.config.update(
    SERVER_NAME="localhost:5000",
    SECRET_KEY="SECRET_KEY",
    # No permission checking
    RECORDS_REST_DEFAULT_READ_PERMISSION_FACTORY=None,
    SQLALCHEMY_TRACK_MODIFICATIONS=True,
    SQLALCHEMY_DATABASE_URI=os.getenv(
        "SQLALCHEMY_DATABASE_URI", "sqlite:///app.db"
    ),
)
app.config["RECORDS_REST_ENDPOINTS"] = CIRCULATION_REST_ENDPOINTS
app.url_map.converters["pid"] = PIDConverter
InvenioDB(app)
InvenioRecords(app)
InvenioREST(app)
InvenioRecordsREST(app)
InvenioPIDStore(app)
InvenioSearch(app)
InvenioJSONSchemas(app)
InvenioCirculation(app)
app.register_blueprint(create_blueprint_from_app(app))
app.register_blueprint(create_loan_actions_blueprint(app))
app.register_blueprint(create_loan_for_item_blueprint(app))
app.register_blueprint(create_loan_replace_item_blueprint(app))


@app.cli.group()
def fixtures():
    """Command for working with test data."""


@fixtures.command()
def loans():
    """Load test data fixture."""
    record_uuid = uuid.uuid4()
    new_loan = {}
    pid = loan_pid_minter(record_uuid, data=new_loan)
    Loan.create(data=new_loan, id_=record_uuid)
    db.session.commit()
    click.secho("Loan #{} created.".format(pid.pid_value), fg="green")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation API."""

from datetime import timedelta


def patron_exists(patron_pid):
    """Return True if patron exists, False otherwise."""
    return True


def item_exists(item_pid):
    """Return True if item exists, False otherwise."""
    return True


def is_item_available(item_pid):
    """Return if item is available for checkout."""
    return True


def can_be_requested(loan):
    """Return True if the given record can be requested, False otherwise."""
    return True


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def item_ref_builder(item_pid):
    """Return the $ref for item_pid."""
    return {"ref": "{}".format(item_pid)}


def item_location_retriever(item_pid):
    """Retrieve the location pid of the passed item pid."""
    return ""


def get_default_loan_duration(loan):
    """Return a default loan duration in number of days."""
    return 30


def get_default_extension_duration(loan):
    """Return a default extension duration in number of days."""
    return 30


def get_default_extension_max_count(loan):
    """Return a default extensions max count."""
    return float("inf")


def is_loan_duration_valid(loan):
    """Validate the loan duration."""
    return loan["end_date"] > loan["start_date"] and loan["end_date"] - loan[
        "start_date"
    ] < timedelta(days=60)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Signals for Invenio-Circulation."""

from blinker import Namespace

_signals = Namespace()

loan_state_changed = _signals.signal('loan-state-changed')
"""Loan state changed signal.

Broadcasted when a loan action is triggered, sending the old and the updated
loan object.
"""

loan_replace_item = _signals.signal('loan-replace-item')
"""Loan item changed signal.

Broadcasted when the item in a Loan is replaced, sending the old and the new
item_pid.
"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

from __future__ import absolute_import, print_function

import json
import uuid
from os.path import dirname, join

import mock
import pytest
from invenio_access import ActionRoles, superuser_access
from invenio_accounts.models import Role
from invenio_app.factory import create_api
from invenio_db import db
from invenio_indexer.api import RecordIndexer
from invenio_records_rest.utils import allow_all
from invenio_search import current_search

from invenio_circulation.api import Loan
from invenio_circulation.permissions import loan_access
from invenio_circulation.pidstore.minters import loan_pid_minter

from .helpers import create_loan, test_views_permissions_factory
from .utils import can_be_requested, get_default_extension_duration, \
    get_default_extension_max_count, get_default_loan_duration, \
    is_item_available, is_loan_duration_valid, item_exists, \
    item_location_retriever, item_ref_builder, patron_exists


@pytest.fixture(scope="module")
def create_app():
    """Return API app."""
    return create_api


@pytest.fixture(scope="module")
def app_config(app_config):
    """Flask application fixture."""
    app_config["JSONSCHEMAS_ENDPOINT"] = "/schema"
    app_config["JSONSCHEMAS_HOST"] = "localhost:5000"
    app_config["RECORDS_REST_DEFAULT_READ_PERMISSION_FACTORY"] = allow_all
    app_config["CIRCULATION_ITEM_EXISTS"] = item_exists
    app_config["CIRCULATION_PATRON_EXISTS"] = patron_exists
    app_config["CIRCULATION_ITEM_REF_BUILDER"] = item_ref_builder
    app_config["CIRCULATION_ITEM_LOCATION_RETRIEVER"] = item_location_retriever
    app_config["CIRCULATION_POLICIES"] = dict(
        checkout=dict(
            duration_default=get_default_loan_duration,
            duration_validate=is_loan_duration_valid,
            item_available=is_item_available,
        ),
        extension=dict(
            from_end_date=True,
            duration_default=get_default_extension_duration,
            max_count=get_default_extension_max_count,
        ),
        request=dict(
            can_be_requested=can_be_requested
        )
    )
    return app_config


@pytest.fixture()
def loan_created(app):
    """Minimal Loan object."""
    record_uuid = uuid.uuid4()
    new_loan = {}
    loan_pid_minter(record_uuid, data=new_loan)
    loan = Loan.create(data=new_loan, id_=record_uuid)
    db.session.commit()
    yield loan


@pytest.fixture()
def params():
    """Params for API REST payload."""
    return dict(
        transaction_user_pid="user_pid",
        patron_pid="patron_pid",
        item_pid="item_pid",
        transaction_location_pid="loc_pid",
        transaction_date="2018-02-01T09:30:00+02:00",
    )


@pytest.fixture()
def json_headers(app):
    """JSON headers."""
    return [
        ("Content-Type", "application/json"),
        ("Accept", "application/json"),
    ]


@pytest.fixture(scope="session")
def test_data():
    """Load test records."""
    path = "data/loans.json"
    with open(join(dirname(__file__), path)) as fp:
        loans = json.load(fp)
    yield loans


@pytest.fixture()
def test_loans(db, test_data):
    """Load test records."""
    loans = []
    for data in test_data:
        loans.append(create_loan(data))
    db.session.commit()
    yield loans


@pytest.fixture()
def indexed_loans(es, test_loans):
    """Get a function to wait for records to be flushed to index."""
    indexer = RecordIndexer()
    for pid, loan in test_loans:
        indexer.index(loan)
    current_search.flush_and_refresh(index="loans")

    yield test_loans

    for pid, loan in test_loans:
        indexer.delete_by_id(loan.id)
    current_search.flush_and_refresh(index="loans")


@pytest.fixture()
def mock_is_item_available():
    """Mock item_available check."""
    path = "invenio_circulation.transitions.base.is_item_available"
    with mock.patch(path) as mock_is_item_available:
        mock_is_item_available.return_value = True
        yield mock_is_item_available


@pytest.fixture()
def users(db, base_app):
    """Create admin, manager and user."""
    base_app.config[
        "CIRCULATION_VIEWS_PERMISSIONS_FACTORY"
    ] = test_views_permissions_factory

    with db.session.begin_nested():
        datastore = base_app.extensions["security"].datastore

        # create users
        manager = datastore.create_user(
            email="manager@test.com", password="123456", active=True
        )
        admin = datastore.create_user(
            email="admin@test.com", password="123456", active=True
        )
        user = datastore.create_user(
            email="user@test.com", password="123456", active=True
        )

        # Give role to admin
        admin_role = Role(name="admin")
        db.session.add(
            ActionRoles(action=superuser_access.value, role=admin_role)
        )
        datastore.add_role_to_user(admin, admin_role)
        # Give role to user
        manager_role = Role(name="manager")
        db.session.add(
            ActionRoles(action=loan_access.value, role=manager_role)
        )
        datastore.add_role_to_user(manager, manager_role)
    db.session.commit()

    return {"admin": admin, "manager": manager, "user": user}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

import pytest
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search

from invenio_circulation.api import get_loan_for_item
from invenio_circulation.errors import MultipleLoansOnItemError

from .helpers import create_loan


def test_api_circulation_item_loan_pending(app, indexed_loans):
    """Test that no Loan is returned for the given Item if only pendings."""
    pending_item_pid = "item_pending_1"
    loan = get_loan_for_item(pending_item_pid)
    assert loan is None


def test_api_circulation_item_no_loan(app, indexed_loans):
    """Test that no Loan is returned for the given Item if no loans."""
    not_loaned_item_pid = "item_not_loaned"
    loan = get_loan_for_item(not_loaned_item_pid)
    assert loan is None


def test_api_circulation_item_on_loan(app, indexed_loans):
    """Test that Loan is returned for the given Item if one active."""
    multiple_loans_pid = "item_multiple_pending_on_loan_7"
    loan = get_loan_for_item(multiple_loans_pid)
    assert loan['state'] == 'ITEM_ON_LOAN'
    assert loan['item_pid'] == multiple_loans_pid


def test_api_circulation_item_not_found(app, indexed_loans):
    """Test API GET call to check circulation item not found."""
    loan = get_loan_for_item("")
    assert loan is None


def test_multiple_active_loans(app, db, indexed_loans):
    """Test that raises if there are multiple active Loans for given Item."""
    multiple_loans_pid = "item_multiple_pending_on_loan_7"

    test_loan_data = {
        "item_pid": "item_multiple_pending_on_loan_7",
        "patron_pid": "2",
        "state": "ITEM_ON_LOAN",
        "transaction_date": "2018-06-26",
        "transaction_location_pid": "loc_pid",
        "transaction_user_pid": "user_pid",
        "start_date": "2018-07-24",
        "end_date": "2018-08-23"
    }

    pid, loan = create_loan(test_loan_data)
    db.session.commit()
    RecordIndexer().index(loan)
    current_search.flush_and_refresh(index="loans")

    with pytest.raises(MultipleLoansOnItemError):
        get_loan_for_item(multiple_loans_pid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Excpetions tests."""

import pytest

from invenio_circulation.errors import InvalidLoanStateError, \
    InvalidPermissionError, ItemNotAvailableError, LoanMaxExtensionError, \
    MissingRequiredParameterError, MultipleLoansOnItemError, \
    NotImplementedConfigurationError, NoValidTransitionAvailableError, \
    RecordCannotBeRequestedError, TransitionConditionsFailedError, \
    TransitionConstraintsViolationError


def test_not_implemented(app):
    """Test NotImplementedConfigurationError."""
    config_variable = 'CONFIG_VAR'
    msg = (
        "Function is not implemented. Implement this function in your module "
        "and pass it to the config variable '{}'".format(config_variable)
    )
    with pytest.raises(NotImplementedConfigurationError) as ex:
        raise NotImplementedConfigurationError(config_variable="CONFIG_VAR")
    assert ex.value.code == 400
    assert ex.value.description == msg


def test_invalid_permission(app):
    """Test InvalidPermissionError."""
    permission = 'simple_user'
    msg = "This action is not permitted for your role '{}'".format(permission)
    with pytest.raises(InvalidPermissionError) as ex:
        raise InvalidPermissionError(permission=permission)
    assert ex.value.code == 403
    assert ex.value.description == msg


def test_record_cannot_be_requested(app):
    """Test record cannot be requested."""
    state = 'loan_state'
    item_pid = 'item_pid'
    msg = (
        'Transition to {0} failed.'
        'Document {1} can not be requested.'
    ).format(state, item_pid)

    with pytest.raises(RecordCannotBeRequestedError) as ex:
        raise RecordCannotBeRequestedError(description=msg)
    assert ex.value.code == 400
    assert ex.value.description == msg


def test_item_not_available(app):
    """Test item not available."""
    item_pid = "1"
    transition = "checkout"
    msg = (
        "The item requested with pid '{0}' is not available. "
        "Transition to '{1}' has failed.".format(item_pid, transition)
    )
    with pytest.raises(ItemNotAvailableError) as ex:
        raise ItemNotAvailableError(item_pid=item_pid, transition=transition)

    assert ex.value.code == 400
    assert ex.value.description == msg


def test_multiple_loans_on_item(app):
    """Test multiple loans on item."""
    item_pid = "1"
    msg = "Multiple active loans on item with pid '{}'".format(item_pid)
    with pytest.raises(MultipleLoansOnItemError) as ex:
        raise MultipleLoansOnItemError(item_pid=item_pid)

    assert ex.value.code == 400
    assert ex.value.description == msg


def test_invalid_loan_state(app):
    """Test invalid loan state."""
    state = "not_valid"
    msg = "Invalid loan state '{}'".format(state)
    with pytest.raises(InvalidLoanStateError) as ex:
        raise InvalidLoanStateError(state=state)
    assert ex.value.code == 400
    assert ex.value.description == msg


def test_no_valid_transitions_available(app):
    """Test no valid transitions available."""
    loan_pid = 'loan_pid'
    state = "not_valid"
    msg = (
        "For the loan with pid '{0}' there are no valid transitions from "
        "its current state '{1}'".format(loan_pid, state)
    )
    with pytest.raises(NoValidTransitionAvailableError) as ex:
        raise NoValidTransitionAvailableError(loan_pid=loan_pid, state=state)
    assert ex.value.code == 400
    assert ex.value.description == msg


def test_transition_conditions_failed(app):
    """Test transition conditions fail."""
    msg = "Transition conditions have failed."

    with pytest.raises(TransitionConditionsFailedError) as ex:
        raise TransitionConditionsFailedError(description=msg)
    assert ex.value.code == 400
    assert ex.value.description == msg


def test_transition_constraints_violation(app):
    """Test transition constraints violation."""
    msg = "The contraints of the transitions have been wildely violated."
    with pytest.raises(TransitionConstraintsViolationError) as ex:
        raise TransitionConstraintsViolationError(description=msg)
    assert ex.value.code == 400
    assert ex.value.description == msg


def test_missing_property(app):
    """Test missing property."""
    msg = "The property pid is missing."
    with pytest.raises(MissingRequiredParameterError) as ex:
        raise MissingRequiredParameterError(description=msg)
    assert ex.value.code == 400
    assert ex.value.description == msg


def test_loan_max_extensions(app):
    """Test loan max extension error."""
    loan_pid = "loan_pid"
    extension_count = 42
    msg = (
        "You have reached the maximum amount of extesions '{}' "
        "for loan '{}'".format(extension_count, loan_pid)
    )
    with pytest.raises(LoanMaxExtensionError) as ex:
        raise LoanMaxExtensionError(
            loan_pid=loan_pid,
            extension_count=extension_count
        )
    assert ex.value.code == 400
    assert ex.value.description == msg



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

import json

from flask import url_for
from invenio_db import db

from invenio_circulation.pidstore.fetchers import loan_pid_fetcher
from invenio_circulation.proxies import current_circulation
from invenio_circulation.views import build_url_action_for_pid


def test_rest_get_loan(app, json_headers, loan_created):
    """Test API GET call to fetch a loan by PID."""
    loan_pid = loan_pid_fetcher(loan_created.id, loan_created)
    expected_links = {
        'actions': {
            'request': build_url_action_for_pid(loan_pid, 'request'),
            'checkout': build_url_action_for_pid(loan_pid, 'checkout')
        }
    }

    with app.test_client() as client:
        url = url_for('invenio_records_rest.loanid_item',
                      pid_value=loan_pid.pid_value)
        res = client.get(url, headers=json_headers)

    assert res.status_code == 200
    loan_dict = json.loads(res.data.decode('utf-8'))
    assert loan_dict['metadata']['state'] == loan_created['state']
    assert loan_dict['links'] == expected_links


def _post(app, json_headers, params, pid_value, action):
    """Perform API POST with the given param."""
    with app.test_client() as client:
        url = url_for('invenio_circulation_loan_actions.loanid_actions',
                      pid_value=pid_value, action=action)
        res = client.post(url, headers=json_headers, data=json.dumps(params))
        payload = json.loads(res.data.decode('utf-8'))
    return res, payload


def test_rest_explicit_loan_valid_action(
    app, json_headers, params, mock_is_item_available, loan_created
):
    """Test API valid action on loan."""
    loan_pid = loan_pid_fetcher(loan_created.id, loan_created)

    res, payload = _post(app, json_headers, params,
                         pid_value=loan_pid.pid_value, action='checkout')
    assert res.status_code == 202
    assert payload['metadata']['state'] == 'ITEM_ON_LOAN'


def test_rest_automatic_loan_valid_action(
    app, json_headers, params, loan_created
):
    """Test API valid action on loan."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(params, trigger='request',
               pickup_location_pid='pickup_location_pid')
    )
    db.session.commit()
    assert loan['state'] == 'PENDING'

    app.config[
        'CIRCULATION_ITEM_LOCATION_RETRIEVER'
    ] = lambda x: 'pickup_location_pid'

    loan_pid = loan_pid_fetcher(loan.id, loan)

    res, payload = _post(app, json_headers, params,
                         pid_value=loan_pid.pid_value, action='next')
    assert res.status_code == 202
    assert payload['metadata']['state'] == 'ITEM_AT_DESK'


def test_rest_loan_invalid_action(
    app, json_headers, params, mock_is_item_available, loan_created
):
    """Test API invalid action on loan."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(params, trigger='request',
               pickup_location_pid='pickup_location_pid')
    )
    db.session.commit()
    assert loan['state'] == 'PENDING'

    loan_pid = loan_pid_fetcher(loan.id, loan)

    res, payload = _post(app, json_headers, params,
                         pid_value=loan_pid.pid_value, action='extend')
    assert res.status_code == 400
    assert 'message' in payload



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for the circulation of bibliographic items."""

from invenio_records_rest.utils import allow_all

from .api import Loan
from .links import loan_links_factory
from .permissions import views_permissions_factory
from .pidstore.pids import _LOANID_CONVERTER, CIRCULATION_LOAN_FETCHER, \
    CIRCULATION_LOAN_MINTER, CIRCULATION_LOAN_PID_TYPE
from .search.api import LoansSearch
from .transitions.transitions import CreatedToPending, \
    ItemAtDeskToItemOnLoan, ItemInTransitHouseToItemReturned, \
    ItemOnLoanToItemInTransitHouse, ItemOnLoanToItemOnLoan, \
    ItemOnLoanToItemReturned, PendingToItemAtDesk, \
    PendingToItemInTransitPickup, ToItemOnLoan
from .utils import can_be_requested, get_default_extension_duration, \
    get_default_extension_max_count, get_default_loan_duration, \
    is_item_available, is_loan_duration_valid, item_exists, \
    item_location_retriever, item_ref_builder, patron_exists

CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT = None
"""Function that returns a list of item PIDs given a Document PID."""

CIRCULATION_DOCUMENT_RETRIEVER_FROM_ITEM = None
"""Function that returns the Document PID of a given Item PID."""

CIRCULATION_STATES_LOAN_ACTIVE = ['ITEM_AT_DESK',
                                  'ITEM_ON_LOAN',
                                  'ITEM_IN_TRANSIT_FOR_PICKUP',
                                  'ITEM_IN_TRANSIT_TO_HOUSE',
                                  ]
"""Defines the list of Loan states for which a Loan is considered as active.

Items that have attached loans with these circulation statuses are
not available to be loaned by patrons."""

CIRCULATION_STATES_LOAN_COMPLETED = ['ITEM_RETURNED']
"""Defines the list of states that a loan is considered completed.

Loans with these states are considered as valid past loans for the item they
refer to."""

CIRCULATION_LOAN_TRANSITIONS_DEFAULT_PERMISSION_FACTORY = allow_all
"""Default permission factory for all Loans transitions."""

CIRCULATION_LOAN_TRANSITIONS = {
    'CREATED': [
        dict(dest='PENDING', trigger='request', transition=CreatedToPending),
        dict(dest='ITEM_ON_LOAN', trigger='checkout', transition=ToItemOnLoan)
    ],
    'PENDING': [
        dict(dest='ITEM_AT_DESK', transition=PendingToItemAtDesk),
        dict(dest='ITEM_IN_TRANSIT_FOR_PICKUP',
             transition=PendingToItemInTransitPickup),
        dict(dest='ITEM_ON_LOAN', trigger='checkout', transition=ToItemOnLoan),
        dict(dest='CANCELLED', trigger='cancel')
    ],
    'ITEM_AT_DESK': [
        dict(dest='ITEM_ON_LOAN', transition=ItemAtDeskToItemOnLoan),
        dict(dest='CANCELLED', trigger='cancel')
    ],
    'ITEM_IN_TRANSIT_FOR_PICKUP': [
        dict(dest='ITEM_AT_DESK'),
        dict(dest='CANCELLED', trigger='cancel')
    ],
    'ITEM_ON_LOAN': [
        dict(dest='ITEM_RETURNED', transition=ItemOnLoanToItemReturned),
        dict(dest='ITEM_IN_TRANSIT_TO_HOUSE',
             transition=ItemOnLoanToItemInTransitHouse),
        dict(dest='ITEM_ON_LOAN', transition=ItemOnLoanToItemOnLoan,
             trigger='extend'),
        dict(dest='CANCELLED', trigger='cancel')
    ],
    'ITEM_IN_TRANSIT_TO_HOUSE': [
        dict(dest='ITEM_RETURNED',
             transition=ItemInTransitHouseToItemReturned),
        dict(dest='CANCELLED', trigger='cancel')
    ],
    'ITEM_RETURNED': [],
    'CANCELLED': [],
}
"""Default circulation Loans transitions."""

CIRCULATION_LOAN_INITIAL_STATE = 'CREATED'
"""Define the initial state name of a Loan."""

CIRCULATION_PATRON_EXISTS = patron_exists
"""Function that returns True if the given Patron exists."""

CIRCULATION_ITEM_EXISTS = item_exists
"""Function that returns True if the given Item exists."""

CIRCULATION_ITEM_REF_BUILDER = item_ref_builder
"""Function that builds $ref to an `Item` record."""

CIRCULATION_ITEM_LOCATION_RETRIEVER = item_location_retriever
"""Function that returns the Location PID of the given Item."""

CIRCULATION_ITEM_RESOLVING_PATH = ""
"""Path to use for the resolving of the item ref."""

CIRCULATION_ITEM_RESOLVER_ENDPOINT = None
"""Flask endpoint function to handle the item resolving."""

CIRCULATION_POLICIES = dict(
    checkout=dict(
        duration_default=get_default_loan_duration,
        duration_validate=is_loan_duration_valid,
        item_available=is_item_available
    ),
    extension=dict(
        from_end_date=True,
        duration_default=get_default_extension_duration,
        max_count=get_default_extension_max_count
    ),
    request=dict(
        can_be_requested=can_be_requested
    ),
)
"""Default circulation policies when performing an action on a Loan."""

CIRCULATION_REST_ENDPOINTS = dict(
    loanid=dict(
        pid_type=CIRCULATION_LOAN_PID_TYPE,
        pid_minter=CIRCULATION_LOAN_MINTER,
        pid_fetcher=CIRCULATION_LOAN_FETCHER,
        search_class=LoansSearch,
        search_type=None,
        record_class=Loan,
        record_serializers={
            'application/json': ('invenio_records_rest.serializers'
                                 ':json_v1_response'),
        },
        search_serializers={
            'application/json': ('invenio_records_rest.serializers'
                                 ':json_v1_search'),
        },
        list_route='/circulation/loans/',
        item_route='/circulation/loans/<{0}:pid_value>'.format(
            _LOANID_CONVERTER),
        default_media_type='application/json',
        max_result_window=10000,
        error_handlers=dict(),
        create_permission_factory_imp=allow_all
    ),
)
"""REST endpoint configuration for circulation APIs."""

CIRCULATION_LOAN_LINKS_FACTORY = loan_links_factory
"""Links factory for Loan record."""

CIRCULATION_VIEWS_PERMISSIONS_FACTORY = views_permissions_factory
"""Permissions factory for circulation views to handle actions."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation API."""


import ciso8601

from .errors import NotImplementedConfigurationError


def patron_exists(patron_pid):
    """Return True if patron exists, False otherwise."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_PATRON_EXISTS"
    )


def item_exists(item_pid):
    """Return True if item exists, False otherwise."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_ITEM_EXISTS"
    )


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def item_ref_builder(loan_pid):
    """Return the $ref for item_pid."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_ITEM_REF_BUILDER"
    )


def item_location_retriever(item_pid):
    """Retrieve the location pid of the passed item pid."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_ITEM_LOCATION_RETRIEVER"
    )


def is_item_available(item_pid):
    """Return if item is available for checkout."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.checkout.is_item_available"
    )


def can_be_requested(loan):
    """Should return True if document or item can be requested."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.request.can_be_requested"
    )


def get_default_loan_duration(loan):
    """Return a default loan duration in number of days."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.checkout.duration_default"
    )


def is_loan_duration_valid(loan):
    """Validate the loan duration."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.checkout.duration_validate"
    )


def get_default_extension_duration(loan):
    """Return a default extension duration in number of days."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.extension.duration_default"
    )


def get_default_extension_max_count(loan):
    """Return a default extensions max count."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.extension.max_count"
    )


def parse_date(str_date):
    """Parse string date with timezone and return a datetime object."""
    return ciso8601.parse_datetime(str_date)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan mandatory constraints."""

import pytest

from invenio_circulation.errors import ItemDoNotMatchError, \
    ItemNotAvailableError, MissingRequiredParameterError, \
    TransitionConstraintsViolationError
from invenio_circulation.pidstore.fetchers import loan_pid_fetcher
from invenio_circulation.proxies import current_circulation

from .helpers import SwappedConfig


def test_should_fail_when_missing_required_params(loan_created):
    """Test that transition fails when there required params are missing."""
    with pytest.raises(MissingRequiredParameterError):
        current_circulation.circulation.trigger(
            loan_created, **dict(patron_pid="pid", trigger="checkout")
        )


def test_should_fail_when_item_not_exist(loan_created, params):
    """Test that transition fails when loan item do not exists."""
    with pytest.raises(ItemNotAvailableError):
        with SwappedConfig("CIRCULATION_ITEM_EXISTS", lambda x: False):
            current_circulation.circulation.trigger(
                loan_created, **dict(params, trigger="checkout")
            )


def test_should_fail_when_item_is_changed(
    loan_created, db, params
):
    """Test that constraints fail if item for an existing loan is changed."""
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request")
        )
    db.session.commit()

    params["item_pid"] = "different_item_pid"
    with pytest.raises(ItemDoNotMatchError):
        current_circulation.circulation.trigger(loan, **dict(params))


def test_should_fail_when_patron_not_exist(loan_created, params):
    """Test that transition fails when loan patron do not exists."""
    with pytest.raises(TransitionConstraintsViolationError):
        with SwappedConfig("CIRCULATION_PATRON_EXISTS", lambda x: False):
            current_circulation.circulation.trigger(
                loan_created, **dict(params, trigger="checkout")
            )


def test_should_fail_when_patron_is_changed(
    loan_created, db, params
):
    """Test that constraints fail if patron for an existing loan is changed."""
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request")
        )
    db.session.commit()

    params["patron_pid"] = "different_patron_pid"
    with pytest.raises(TransitionConstraintsViolationError):
        current_circulation.circulation.trigger(loan, **dict(params))


def test_persist_loan_parameters(
    loan_created, db, params, mock_is_item_available
):
    """Test that input params are correctly persisted."""
    loan_pid = loan_pid_fetcher(loan_created.id, loan_created)
    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    db.session.commit()

    params["loan_pid"] = loan_pid.pid_value
    params["state"] = "ITEM_ON_LOAN"
    params["start_date"] = loan["start_date"]
    params["end_date"] = loan["end_date"]
    params["trigger"] = loan["trigger"]
    params["$schema"] = "https://localhost:5000/schema/loans/loan-v1.0.0.json"
    assert loan == params



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.\

"""."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan JSON schema."""
from copy import deepcopy
from json import loads

from pkg_resources import resource_string

from invenio_circulation.proxies import current_circulation

schema_in_bytes = resource_string(
    'invenio_circulation.schemas', 'loans/loan-v1.0.0.json'
)


def test_state_enum(app):
    """Test that schema Loan states corresponds to transitions states."""
    schema = loads(schema_in_bytes.decode('utf8'))
    state_dict = schema.get('properties').get('state')
    assert 'enum' in state_dict
    all_states = app.config['CIRCULATION_LOAN_TRANSITIONS'].keys()
    assert not (set(state_dict.get('enum')) - set(all_states))


def test_state_checkout_with_loan_pid(
    loan_created, db, params, mock_is_item_available
):
    """Test that created Loan validates after a checkout action."""
    new_params = deepcopy(params)
    new_params['trigger'] = 'checkout'
    loan = current_circulation.circulation.trigger(loan_created, **new_params)
    loan.validate()


def test_indexed_loans(indexed_loans):
    """Test mappings, index creation and loans indexing."""
    assert indexed_loans



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for the circulation of bibliographic items."""

from __future__ import absolute_import, print_function

from copy import deepcopy

from flask import current_app
from invenio_indexer.api import RecordIndexer
from invenio_records_rest.utils import obj_or_import_string
from werkzeug.utils import cached_property

from . import config
from .errors import InvalidLoanStateError, NoValidTransitionAvailableError, \
    TransitionConditionsFailedError
from .pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from .search.api import LoansSearch
from .transitions.base import Transition


class InvenioCirculation(object):
    """Invenio-Circulation extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.app = app
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.config.setdefault("RECORDS_REST_ENDPOINTS", {})

        app.config["CIRCULATION_REST_ENDPOINTS"].setdefault("loanid", {})[
            "links_factory_imp"
        ] = app.config["CIRCULATION_LOAN_LINKS_FACTORY"]

        app.config["RECORDS_REST_ENDPOINTS"].update(
            app.config["CIRCULATION_REST_ENDPOINTS"]
        )
        app.extensions["invenio-circulation"] = self

    def init_config(self, app):
        """Initialize configuration."""
        app.config.setdefault(
            "CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT", lambda x: []
        )
        app.config.setdefault(
            "CIRCULATION_DOCUMENT_RETRIEVER_FROM_ITEM", lambda x: None
        )
        for k in dir(config):
            if k.startswith("CIRCULATION_"):
                app.config.setdefault(k, getattr(config, k))

    @cached_property
    def circulation(self):
        """Return the Circulation state machine."""
        return _Circulation(
            transitions_config=deepcopy(
                current_app.config["CIRCULATION_LOAN_TRANSITIONS"]
            )
        )

    def _get_endpoint_config(self):
        """Return endpoint configuration for circulation."""
        endpoints = self.app.config.get('CIRCULATION_REST_ENDPOINTS', [])
        pid_type = CIRCULATION_LOAN_PID_TYPE
        return endpoints.get(pid_type, {})

    @cached_property
    def loan_search(self):
        """Return the current Loan search instance."""
        circ_endpoint = self._get_endpoint_config()
        _cls = circ_endpoint.get('search_class', LoansSearch)
        return obj_or_import_string(_cls)()

    @cached_property
    def loan_indexer(self):
        """Return the current Loan indexer instance."""
        circ_endpoint = self._get_endpoint_config()
        _cls = circ_endpoint.get('indexer_class', RecordIndexer)
        return obj_or_import_string(_cls)()


class _Circulation(object):
    """Circulation state sachine."""

    def __init__(self, transitions_config):
        """Constructor."""
        self.transitions = {}
        for src_state, transitions in transitions_config.items():
            self.transitions.setdefault(src_state, [])
            for t in transitions:
                _cls = t.pop("transition", Transition)
                instance = _cls(**dict(t, src=src_state))
                self.transitions[src_state].append(instance)

    def _validate_current_state(self, state):
        """Validate that the given loan state is configured."""
        if not state or state not in self.transitions:
            raise InvalidLoanStateError(state=state)

    def trigger(self, loan, **kwargs):
        """Trigger the action to transit a Loan to the next state."""
        current_state = loan.get("state")
        self._validate_current_state(current_state)

        for t in self.transitions[current_state]:
            try:
                t.execute(loan, **kwargs)
                return loan
            except TransitionConditionsFailedError as ex:
                current_app.logger.exception(ex)
                pass

        raise NoValidTransitionAvailableError(
            loan_pid=str(loan.id), state=current_state
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation views."""

from copy import deepcopy

from flask import Blueprint, current_app, jsonify, request, url_for
from invenio_db import db
from invenio_records_rest.utils import obj_or_import_string
from invenio_records_rest.views import pass_record
from invenio_rest import ContentNegotiatedMethodView

from .api import get_loan_for_item
from .errors import InvalidLoanStateError, ItemNotAvailableError, \
    MissingRequiredParameterError
from .permissions import need_permissions
from .pidstore.fetchers import loan_pid_fetcher
from .pidstore.pids import _LOANID_CONVERTER, CIRCULATION_LOAN_PID_TYPE
from .proxies import current_circulation
from .signals import loan_replace_item


def extract_transitions_from_app(app):
    """Return all possible actions for configured transitions."""
    transitions_config = app.config.get("CIRCULATION_LOAN_TRANSITIONS", {})
    distinct_actions = set()
    for src_state, transitions in transitions_config.items():
        for t in transitions:
            distinct_actions.add(t.get("trigger", "next"))
    return distinct_actions


def build_url_action_for_pid(pid, action):
    """Build urls for Loan actions."""
    return url_for(
        "invenio_circulation_loan_actions.{0}_actions".format(pid.pid_type),
        pid_value=pid.pid_value,
        action=action,
        _external=True,
    )


def create_loan_actions_blueprint(app):
    """Create a blueprint for Loan actions."""
    blueprint = Blueprint(
        "invenio_circulation_loan_actions", __name__, url_prefix=""
    )

    endpoints = app.config.get("CIRCULATION_REST_ENDPOINTS", [])
    pid_type = CIRCULATION_LOAN_PID_TYPE
    options = endpoints.get(pid_type, {})
    if options:
        options = deepcopy(options)
        serializers = {}
        if "record_serializers" in options:
            rec_serializers = options.get("record_serializers")
            serializers = {
                mime: obj_or_import_string(func)
                for mime, func in rec_serializers.items()
            }

        loan_actions = LoanActionResource.as_view(
            LoanActionResource.view_name.format(pid_type),
            serializers=serializers,
            ctx={},
        )

        distinct_actions = extract_transitions_from_app(app)
        url = "{0}/<any({1}):action>".format(
            options["item_route"], ",".join(distinct_actions)
        )

        blueprint.add_url_rule(url, view_func=loan_actions, methods=["POST"])

    return blueprint


class LoanActionResource(ContentNegotiatedMethodView):
    """Loan action resource."""

    view_name = "{0}_actions"

    def __init__(self, serializers, ctx, *args, **kwargs):
        """Constructor."""
        super(LoanActionResource, self).__init__(serializers, *args, **kwargs)
        for key, value in ctx.items():
            setattr(self, key, value)

    @need_permissions("loan-actions")
    @pass_record
    def post(self, pid, record, action, **kwargs):
        """Handle loan action."""
        params = request.get_json()
        record = current_circulation.circulation.trigger(
            record, **dict(params, trigger=action)
        )
        db.session.commit()
        return self.make_response(
            pid,
            record,
            202,
            links_factory=current_app.config.get(
                "CIRCULATION_LOAN_LINKS_FACTORY"
            ),
        )


def create_loan_for_item_blueprint(app):
    """Create a blueprint for Loan status of Items."""
    blueprint = Blueprint(
        "invenio_circulation_loan_for_item", __name__, url_prefix=""
    )

    rec_serializers = {
        "application/json": (
            "invenio_records_rest.serializers" ":json_v1_response"
        )
    }
    serializers = {
        mime: obj_or_import_string(func)
        for mime, func in rec_serializers.items()
    }

    loan_request = ItemLoanResource.as_view(
        ItemLoanResource.view_name, serializers=serializers, ctx={}
    )

    url = "circulation/items/<pid_value>/loan"

    blueprint.add_url_rule(url, view_func=loan_request, methods=["GET"])
    return blueprint


class ItemLoanResource(ContentNegotiatedMethodView):
    """Item circulation state resource."""

    view_name = "loan_resource"

    def __init__(self, serializers, ctx, *args, **kwargs):
        """Resource view constructor."""
        super(ItemLoanResource, self).__init__(serializers, *args, **kwargs)
        for key, value in ctx.items():
            setattr(self, key, value)

    @need_permissions("loan-read-access")
    def get(self, *args, **kwargs):
        """Handle GET request for item state."""
        item_pid = kwargs.get("pid_value", None)
        if not item_pid:
            raise ItemNotAvailableError(item_pid=item_pid)

        loan = get_loan_for_item(item_pid)
        if loan:
            loan_pid = loan_pid_fetcher(loan.id, loan)
            return self.make_response(
                loan_pid,
                loan,
                200,
                links_factory=current_app.config.get(
                    "CIRCULATION_LOAN_LINKS_FACTORY"
                ),
            )
        return jsonify({})


def create_loan_replace_item_blueprint(app):
    """Create a blueprint for replacing Loan Item."""
    blueprint = Blueprint(
        "invenio_circulation_loan_replace_item", __name__, url_prefix=""
    )

    rec_serializers = {
        "application/json": (
            "invenio_records_rest.serializers" ":json_v1_response"
        )
    }
    serializers = {
        mime: obj_or_import_string(func)
        for mime, func in rec_serializers.items()
    }
    replace_item_view = LoanReplaceItemResource.as_view(
        LoanReplaceItemResource.view_name.format(CIRCULATION_LOAN_PID_TYPE),
        serializers=serializers,
        ctx={},
    )

    url = "circulation/loans/<{0}:pid_value>/replace-item".format(
        _LOANID_CONVERTER
    )
    blueprint.add_url_rule(url, view_func=replace_item_view, methods=["POST"])
    return blueprint


def validate_replace_item(loan, data):
    """Validate data before replacing an item in loan."""
    active_states = current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
    if loan["state"] not in active_states:
        raise InvalidLoanStateError(description=(
            "Cannot replace item in a loan that is not in active state. "
            "Current loan state '{}'".format(loan["state"])
        ))

    item_pid = data.get("item_pid")
    if not item_pid:
        raise MissingRequiredParameterError(
            description="Parameter item_pid is required."
        )

    item_exists = current_app.config.get("CIRCULATION_ITEM_EXISTS")(item_pid)
    if not item_exists:
        raise ItemNotAvailableError(item_pid=item_pid)


def loan_update(loan, data):
    """Update loan with new properties."""
    item_ref_builder = current_app.config.get("CIRCULATION_ITEM_REF_BUILDER")
    item_pid = data.get("item_pid")
    loan["item_pid"] = item_pid
    loan["item"] = item_ref_builder(item_pid)
    return loan


class LoanReplaceItemResource(ContentNegotiatedMethodView):
    """Loan update resource."""

    view_name = "loan_replace_item_resource"

    def __init__(self, serializers, ctx, *args, **kwargs):
        """Constructor."""
        super(LoanReplaceItemResource, self).__init__(
            serializers, *args, **kwargs
        )
        for key, value in ctx.items():
            setattr(self, key, value)

    @need_permissions("loan-actions")
    @pass_record
    def post(self, pid, record, *args, **kwargs):
        """Handle POST request to update loan with new item."""
        old_item_pid = record["item_pid"]
        data = request.get_json()
        validate_replace_item(record, data)
        record = loan_update(record, data)
        db.session.commit()

        current_circulation.loan_indexer.index(record)
        if old_item_pid:
            loan_replace_item.send(
                self,
                old_item_pid=old_item_pid,
                new_item_pid=record["item_pid"]
            )

        return self.make_response(
            pid,
            record,
            202,
            links_factory=current_app.config.get(
                "CIRCULATION_LOAN_LINKS_FACTORY"
            ),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation transitions conditions."""

from flask import current_app


def is_same_location(item_pid, input_location_pid):
    """Return True if item belonging location is same as input parameter."""
    item_location_pid = current_app.config.get(
        'CIRCULATION_ITEM_LOCATION_RETRIEVER'
    )(item_pid)

    return input_location_pid == item_location_pid
